//Tcp.cc

/* Author:  Marty Kraimer Date:    08MAY1998 */
 
/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include <ioLib.h>
#include <sockLib.h>
#include <inetLib.h>
#include <netinet/tcp.h>
//inet_addr is incorrectly defined in inetLib
#include "TcpBadProto.h"
#include <errno.h>

#include <vxWorks.h>
#include <taskLib.h>

#include "mpfType.h"
#include "WatchDog.h"
#include "Tcp.h"

int tcpDebug=0;
static const int nConnectTry = 5;
static const int connectRetryDelay = 600;


//The following is for exclusive use by this source module
// Tcp is an interface class that uses TcpPrivate for implementation
class TcpPrivate {
public:
    TcpPrivate(bool isServer,int port,const char *address,int bufferSize);
    ~TcpPrivate();
    void connect(void *userPvt, tcpConnectCallback ,tcpReceiveCallback);
    int send(void *data,int32 size);
    int flush();
    int sendFlush(void *data,int32 size);
    //following are for use by this source module
    void lock() {semTake(lockSem,WAIT_FOREVER);}
    void unlock() {semGive(lockSem);}
    void issueError(const char *message);
    void issueErrorExit(const char *message);
    void dumpData(const char *who,bool isServer,void *data, int32 nbytes);
    void setSocketOptions(int fd);
    void *get(int32 nbytes);
    void receiveMessages();
    int initSocket();
    int connectServer();
    int connectClient();
    //beginning of data
    tcpConnectCallback connectCallback;
    tcpReceiveCallback receiveCallback;
    void *userPvt;
    int port;
    const int bufferSize;
    bool connected;
    bool isServer;
    int taskid;
    int32 outbufOffset;
    int32 inbufReceiveBytes;
    int32 inbufOffset;
    char *address;
    char *inbuf;
    char *outbuf;
    int sockfd;
    int clientfd;
    struct sockaddr_in servaddr;
    struct sockaddr_in clientaddr; // Used by server only
    SEM_ID lockSem;
    WatchDog *wdId;
    // stat fields
    int nSend;
    int nReceive;
    int nTcpSend;
    int nTcpReceive;
    int sendPerSecond;
    int receivePerSecond;
    int tcpSendPerSecond;
    int tcpReceivePerSecond;
private:
    //disallow following
    TcpPrivate();
    TcpPrivate(const TcpPrivate &);
    TcpPrivate & operator=(const TcpPrivate &);
};

void tcpTask(TcpPrivate *ptcp);


static int secondsBetweenCalc = 10;
static void calculateStats(void *pvt)
{
    TcpPrivate *ptcp = (TcpPrivate *)pvt;
    int nSend;
    int nReceive;
    int nTcpSend;
    int nTcpReceive;

    nSend = ptcp->nSend; ptcp->nSend = 0;
    nReceive = ptcp->nReceive; ptcp->nReceive = 0;
    nTcpSend = ptcp->nTcpSend; ptcp->nTcpSend = 0;
    nTcpReceive = ptcp->nTcpReceive; ptcp->nTcpReceive = 0;
    ptcp->sendPerSecond = nSend/secondsBetweenCalc;
    ptcp->receivePerSecond = nReceive/secondsBetweenCalc;
    ptcp->tcpSendPerSecond = nTcpSend/secondsBetweenCalc;
    ptcp->tcpReceivePerSecond = nTcpReceive/secondsBetweenCalc;
    ptcp->wdId->start(secondsBetweenCalc,calculateStats,(void *)ptcp);
}


Tcp::Tcp(bool isServer,int port,const char *address,int bufferSize)
{ tcpPvt = new TcpPrivate(isServer,port,address,bufferSize);}

Tcp::~Tcp() {delete (TcpPrivate *)tcpPvt;}

void Tcp::connect(void *userPvt, tcpConnectCallback connectCallback,
    tcpReceiveCallback receiveCallback)
{
    TcpPrivate *pvt = (TcpPrivate *)tcpPvt;
    pvt->connect(userPvt,connectCallback,receiveCallback);
    return;
}

int Tcp::send(void *data,int32 size)
{
    TcpPrivate *pvt = (TcpPrivate *)tcpPvt;
    pvt->lock();
    int status = pvt->send(data,size);
    pvt->unlock();
    return(status);
}

int Tcp::flush()
{
    TcpPrivate *pvt = (TcpPrivate *)tcpPvt;
    pvt->lock();
    int status = pvt->flush();
    pvt->unlock();
    return(status);
}

int Tcp::sendFlush(void *data,int32 size)
{
    TcpPrivate *pvt = (TcpPrivate *)tcpPvt;
    pvt->lock();
    int status = pvt->sendFlush(data,size);
    pvt->unlock();
    return(status);
}

void Tcp::printStats()
{
    TcpPrivate *pvt = (TcpPrivate *)tcpPvt;
    printf("   sendPerSec %d receivePerSec %d",
	pvt->sendPerSecond, pvt->receivePerSecond);
    printf("  tcpSendPerSec %d tcpReceivePerSec %d\n",
	pvt->tcpSendPerSecond, pvt->tcpReceivePerSecond);
}

TcpPrivate::TcpPrivate(bool isServer,int port,const char *addr,int bufsize)
    : connectCallback(0), receiveCallback(0), userPvt(0),
    port(port), bufferSize(bufsize), connected(false), isServer(isServer), 
    outbufOffset(0), inbufReceiveBytes(0), inbufOffset(0),
    nSend(0), nReceive(0), nTcpSend(0), nTcpReceive(0),
    sendPerSecond(0), receivePerSecond(0),
    tcpSendPerSecond(0), tcpReceivePerSecond(0)
{
    if(bufferSize<=10) {
	printf("TcpPrivate::TcpPrivate illegal bufferSize %d\n",bufferSize);
	return;
    }
    lockSem = semMCreate(SEM_DELETE_SAFE|SEM_INVERSION_SAFE|SEM_Q_PRIORITY);
    if(!lockSem) {
        printf("Tcp semMCreate failed\n");
        return;
    }
    address = new char[::strlen(addr)+1];
    inbuf = new char[bufferSize];
    outbuf = new char[bufferSize];
    ::strcpy(address,addr);
    wdId = new WatchDog;
    wdId->start(secondsBetweenCalc, calculateStats,(void *)this);
}

TcpPrivate::~TcpPrivate()
{
    if(sockfd>0) {
	int status = ::close(sockfd);
	if(status) issueError("closing sockfd");
    }
    if(clientfd>0) {
	int status = ::close(clientfd);
	if(status) issueError("closing sockfd");
    }
    delete [] outbuf;
    delete [] inbuf;
    delete [] address;
    delete wdId;
    semDelete(lockSem);
}

static const char *TcpClient = "TcpClient";
static const char *TcpServer = "TcpServer";

void TcpPrivate::connect(void *pvt,
tcpConnectCallback connectcallback,tcpReceiveCallback receivecallback)
{
    char *name;
    userPvt = pvt;
    connectCallback = connectcallback;
    receiveCallback = receivecallback;
    if(isServer) {
	name = new char[::strlen(TcpServer) + 1];
	::strcpy(name,TcpServer);
    } else {
	name = new char[::strlen(TcpClient) + 1];
	::strcpy(name,TcpClient);
    }
    taskid = ::taskSpawn(name, 60,VX_FP_TASK,5000,
	(FUNCPTR)tcpTask, (int)this,0,0,0,0,0,0,0,0,0);
    if(taskid==ERROR) {
	issueErrorExit("Tcp::connect taskSpawn error");
    }
}

int TcpPrivate::send(void *data,int32 size)
{
    nSend++;
    int status = 0;
    if(tcpDebug) dumpData("Tcp::send",isServer,data,size);
    if(!connected) return(-1);
    int32 nbytes = htonl(size);
    int nbytesTotal = sizeof(nbytes) + size;
    int nbytesRemaining = bufferSize - outbufOffset;
    if(nbytesTotal>nbytesRemaining) {
	status = flush();
	if(status) return(status);
        if(nbytesTotal>bufferSize) {
            printf("ERROR: message size %d is larger than tcp buffer size %d\n",
                nbytesTotal,bufferSize);
            return(-1);
        }
    }
    ::memcpy(&outbuf[outbufOffset],&nbytes,sizeof(nbytes));
    outbufOffset += sizeof(size);
    ::memcpy(&outbuf[outbufOffset],data,size);
    outbufOffset += size;
    return(0);
}

int TcpPrivate::flush()
{
    if(!connected) {
	outbufOffset = 0;
	return(-1);
    }
    int status = 0;
    int fd = (isServer? clientfd : sockfd);
    if(outbufOffset>0) {
        nTcpSend++;
	if(tcpDebug) {
	    printf("flush nbytes %d ",outbufOffset);
	    long *pdata = (long *)outbuf;
	    for(int i=0; i< (outbufOffset+3)/4; i++) {
		printf(" %8.8lx",*pdata);
		pdata++;
	    }
	    printf("\n");
	}
	int nbytes = ::send(fd,(char *)outbuf,outbufOffset,0);
	if(nbytes != outbufOffset)  {
	    status = -1;
	    printf("Tcp::send (%d,%d) bytes (requested,sent)\n",
		outbufOffset,nbytes);
	}
    }
    outbufOffset = 0;
    return(status);
}

int TcpPrivate::sendFlush(void *data,int32 size)
{
    int status = send(data,size);
    if(status==0) status = flush();
    return(status);
}

void TcpPrivate::issueError(const char *message)
{
    fprintf(stderr,"Tcp error: %s errno %s\n",message,strerror(errno));
}

void TcpPrivate::issueErrorExit(const char *message)
{
    issueError(message);
    taskSuspend(0);
}


void TcpPrivate::dumpData(const char *who,bool isServer,void *data, int32 nbytes)
{
    printf("%s",who);
    if(isServer) printf(" server "); else printf(" client ");
    printf(" nbytes %d ",nbytes);
    int32 *p = (int32 *)data;
    for(int i=0; i<(nbytes+3)/4; i++) printf(" %8.8x ",*p++);
    printf("\n");
}

void TcpPrivate::setSocketOptions(int fd)
{
    int trueValue = 1;
    int status = ::setsockopt(fd,IPPROTO_TCP,TCP_NODELAY,
	(char *)&trueValue, sizeof(trueValue));
    if(status!=0) {
	issueError("Tcp: setsockopt error");
	return;
    }
    int size = bufferSize;
    status = ::setsockopt(fd,SOL_SOCKET,SO_SNDBUF,
	(char *)&size,sizeof(size));
    if(status!=0) {
	issueError("Tcp: setsockopt error");
	return;
    }
    status = ::setsockopt(fd,SOL_SOCKET,SO_RCVBUF,
	(char *)&size,sizeof(size));
    if(status!=0) {
	issueError("Tcp: setsockopt error");
	return;
    }
    return;
}

int TcpPrivate::initSocket()
{
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if(sockfd < 0) {
	issueError("Tcp: socket error");
	return(-1);
    }
    setSocketOptions(sockfd);
    memset(&servaddr,0,sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    if(address)
        servaddr.sin_addr.s_addr = ::inet_addrReplace(address);
    else
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(port);
    return(0);
}

int TcpPrivate::connectServer()
{
    if(initSocket()) return(-1);
    for(int i=0; i<nConnectTry; i++) {
        int rtn = ::bind(sockfd,
	    (struct sockaddr *)&servaddr,sizeof(servaddr));
	if(rtn==0) break;
	issueError("tcpTask bind error");
	::taskDelay(connectRetryDelay);
    }
    for(int i=0; i<nConnectTry; i++) {
        int rtn = ::listen(sockfd,1);
	if(rtn==0) break;
	issueError("tcpTask listen error");
	::taskDelay(connectRetryDelay);
    }
    int size = (int)sizeof(servaddr);
    for(int i=0; i<nConnectTry; i++) {
        clientfd = ::accept(sockfd,(struct sockaddr *)&clientaddr,&size);
	if(clientfd>0) break;
	issueError("tcpTask accept error");
	::taskDelay(connectRetryDelay);
    }
    setSocketOptions(clientfd);
    outbufOffset = 0;
    connected = true;
    return(0);
}

int TcpPrivate::connectClient()
{
    int rtn;
    bool firstTime = true;
    while(true) {
        if(initSocket()) return(-1);
	rtn = ::connect(sockfd,(struct sockaddr *)&servaddr,sizeof(servaddr));
	if(rtn==0) break;
	if(firstTime) {
	    issueError("TcpPrivate::connectClient connect error");
	    firstTime = false;
	}
	rtn = ::close(sockfd);
	if(rtn) issueError("cpPrivate::connectClient close error");
	taskDelay(240);
    }
    outbufOffset = 0;
    connected = true;
    return(0);
}

void *TcpPrivate::get(int32 nbytes)
{
    int fd = (isServer ? clientfd : sockfd);
    if(nbytes<=(inbufReceiveBytes-inbufOffset)) {
	void *rtn = (void *)&inbuf[inbufOffset];
	inbufOffset += nbytes;
	return(rtn);
    }
    if(nbytes>bufferSize) {
	printf("TcpPrivate::get request %d but bufferSize %d\n",
	    nbytes,bufferSize);
	    return(0);
    }
    int32 nbytesRemaining = (inbufReceiveBytes - inbufOffset);
    if(nbytesRemaining<=0) {
	inbufReceiveBytes = 0;
    } else {
	for(int i=0; i< nbytesRemaining; i++) {
	    inbuf[i] = inbuf[i+inbufOffset];
	}
	inbufReceiveBytes = nbytesRemaining;
    }
    inbufOffset = 0;
    while(inbufReceiveBytes<nbytes)  {
        size_t bufSpace = bufferSize - inbufReceiveBytes;
        int32 nnew = ::recv(fd,&inbuf[inbufReceiveBytes],bufSpace,0);
	if(tcpDebug) {
	    printf("recv nbytes %d ",nnew);
	    long *pdata = (long *)&inbuf[inbufReceiveBytes];
	    for(int i=0; i< (nnew+3)/4; i++) {
		printf(" %8.8lx",*pdata);
		pdata++;
	    }
	    printf("\n");
	}
	inbufReceiveBytes += nnew;
	nTcpReceive++;
	if(nnew<0) {
	    issueError("tcpTask recv error");
	    return(0);
	}
	if(nnew==0) {
	    issueError("tcpTask recv received 0 length message");
	    return(0);
	}
    }
    inbufOffset += nbytes;
    return((void *)inbuf);
}

void TcpPrivate::receiveMessages()
{
    (*connectCallback)(userPvt,tcpConnectYes);
    inbufReceiveBytes = 0; inbufOffset =0;
    while(true) {
	void *data;
	int32 value;
	if(!(data = get(sizeof(value)))) break;
	::memcpy(&value,data,sizeof(value));
	int32 messageLength = ntohl(value);
	void *message;
	if(!(message = get(messageLength))) break;
	nReceive++;
	if(tcpDebug) dumpData("tcpTask",isServer,message,messageLength);
        receiveCallback(userPvt,message,messageLength);
    }
}

void tcpTask(TcpPrivate *pvt)
{
    while(true) {
	pvt->sockfd = -1;
	pvt->clientfd = -1;
        if(pvt->isServer) {
	    pvt->connectServer();
	} else {
	    pvt->connectClient();
	}
	if(pvt->connected) pvt->receiveMessages();
	pvt->connected = false;
	if(pvt->sockfd>0) {
	    int status = ::close(pvt->sockfd);
	    if(status) pvt->issueError("closing sockfd");
	}
	if(pvt->clientfd>0) {
	    int status = ::close(pvt->clientfd);
	    if(status) pvt->issueError("closing clientfd");
	}
        (*pvt->connectCallback)(pvt->userPvt,tcpConnectNo);
	taskDelay(connectRetryDelay);
    }
}
